from .services.apigateway import APIGateway
from .services.apigatewayv2 import APIGatewayV2
from .services.cloudformation import CloudFormation
from .services.cloudwatch import CloudWatch
from .services.cloudwatchlogs import CloudWatchLogs
from .services.dynamodb import DynamoDB, dynamodb_type_to_python_type, python_type_to_dynamodb_type
from .services.eventbridge import EventBridge
from .services.iot import IoT
from .services.iam import IAM
from .services.identitystore import IdentityStore
from .services.lambda_aws import Lambda
from .services.resourcegroupstaggpingapi import ResourceGroupsTaggingAPI
from .services.s3 import S3
from .services.secretsmanager import SecretsManager
from .services.sns import SNS
from .services.ssm import SSM
from .services.sqs import SQS
from .services.sso import SSO
from .services.ssoadmin import SSOAdmin
from .services.ssooidc import SSO_OIDC
from .services.sts import STS


from ..utils.common import remove_none_values
from . import AWSService


class SSM(AWSService):
    __servicename__ = 'ssm'

    def get_parameter(self, name, with_decryption = None):
        req_params = remove_none_values({
            "Name":name, 
            "WithDecryption":with_decryption
        })
        
        return self.client.get_parameter(**req_params).get('Parameter').get('Value')
    
    def put_parameter(self, name, value, type, description, overwrite = True):
        req_params = remove_none_values({
            "Name":name,
            "Description":description,
            "Value":value,
            "Type":type,
            "Overwrite":overwrite
        })

        return self.client.put_parameter(**req_params)


from . import AWSService


class SecretsManager(AWSService):
    __servicename__ = 'secretsmanager'

    def get_secret_value(self, secret_id):
        return self.client.get_secret_value(SecretId=secret_id)


import time
from typing import Union
from . import AWSService
from ..utils.common import remove_none_values
from ..utils.logger import get_logger

logger = get_logger('CloudWatchLogs')

class CloudWatchLogs(AWSService):
    __servicename__ = 'logs'

    def get_log_groups(self, next_token = None):
        request_params = remove_none_values({
            'nextToken':next_token
        })
        return self.client.describe_log_groups(**request_params)
    
    def describe_log_groups_with_paginator(self):
        return self.get_result_from_paginator('describe_log_groups', 'logGroups')
    
    def query(self, log_group_names:Union[str, list], query_string, start_time, end_time):
        results, _ = self.query_with_statistics(log_group_names, query_string, start_time, end_time)
        return results
    
    def query_with_statistics(self, log_group_names:Union[str, list], query_string:str, start_time, end_time):
        if isinstance(log_group_names,str):
            log_group_names = [log_group_names]

        query_id = self.start_query(log_group_names, query_string, start_time, end_time)

        results = None # check 
        while True:
            status, results, statistics = self.get_query_results(query_id)
            if status == 'Complete':
                break
            elif status in ['Failed','Cancelled','Timeout','Unknown']:
                logger.error(f'Query execution aborted with status: {status}') 
                break

            time.sleep(1)
        
        returned_values = []
        if results is not None:
            for r in results:
                converted_result = {}
                for f in r:
                    converted_result[f['field']] = f['value']
                returned_values.append(converted_result)
        return returned_values, statistics
    
    def start_query(self, log_group_names, query_string, start_time, end_time):
        response = self.client.start_query(
            logGroupNames = log_group_names,
            queryString = query_string,
            startTime = int(start_time.timestamp() * 1000),
            endTime = int(end_time.timestamp() * 1000)
        )

        return response['queryId']
    
    def get_query_results(self, query_id):
        query_result = self.client.get_query_results(
            queryId = query_id
        )

        return query_result.get('status'), query_result.get('results'), query_result.get('statistics')
    
    def get_log_record(self, log_record_pointer):
        return self.client.get_log_record(logRecordPointer = log_record_pointer)
    
    def start_live_tail(self, log_group_identifiers, callback_func, log_event_filter_pattern = None):
        request_params = remove_none_values({
            'logGroupIdentifiers':log_group_identifiers,
            'logEventFilterPattern':log_event_filter_pattern
        })
        response = self.client.start_live_tail(request_params)

        event_streams = response['responseStream']
        for stream in event_streams:
            callback_func(stream)

    def get_log_group_fields(self, log_group_name):
        return self.client.get_log_group_fields(logGroupName = log_group_name)


import json
from typing import Dict, Union
from . import AWSService
from ..utils.common import remove_none_values


class SNS(AWSService):
    __servicename__ = 'sns'

    def get_topics(self, next_token = None):
        request_params = remove_none_values({
            'NextToken':next_token
        })
        return self.client.list_topics(**request_params)
    
    def list_topics_with_paginator(self):
        return self.get_result_from_paginator('list_topics', 'Topics')
    
    def get_topic_attributes(self, topic_arn):
        response = self.client.get_topic_attributes(TopicArn=topic_arn)
        return response.get('Attributes')
    
    def publish(self, topic_arn, message: Union[dict, str], subject = None, message_attributes:Dict[str, Union[str, bytes]] = None):
        if isinstance(message, dict):
            message = json.dumps(message)
        elif not isinstance(message, str):
            message = str(message)

        msg_attrs = {}
        if message_attributes is not None:
            for k,v in message_attributes.items():
                if isinstance(v, str):
                    msg_attrs[k] = {"DataType": "String", "StringValue": v}
                elif isinstance(v, str):
                    msg_attrs[k] = {"DataType": "Binary", "BinaryValue": v}
                else:
                    msg_attrs[k] = {"DataType": "String", "StringValue": str(v)}

        req_params = remove_none_values({
            "TopicArn": topic_arn,
            "Message": message,
            "Subject": subject,
            "MessageAttributes": msg_attrs
        })

        return self.client.publish(**req_params)


from . import AWSService
from ..utils.common import remove_none_values

class APIGateway(AWSService):
    __servicename__ = 'apigateway'

    def get_apis(self, position = None):
        request_params = remove_none_values({
            'position':position
        })
        return self.client.get_rest_apis(**request_params)
    
    def get_rest_apis_with_paginator(self):
        return self.get_result_from_paginator('get_rest_apis', 'items')
    
    def get_api_resources(self, rest_api_id, position = None):
        request_params = remove_none_values({
            'position':position,
            'restApiId':rest_api_id,
            'embed':["methods"]
        })
        return self.client.get_resources(**request_params)
    
    def get_resources_with_paginator(self, rest_api_id):
        return self.get_result_from_paginator('get_resources', 'items', restApiId=rest_api_id)

    def get_domain_names(self, position = None, limit = 500):
        request_params = remove_none_values({
            'position':position,
            'limit':limit
        })
        return self.client.get_domain_names(**request_params)
    
    def get_domain_names_with_paginator(self):
        return self.get_result_from_paginator('get_domain_names', 'items')
    
    def get_base_path_mappings_by_domain_name(self, domain_name, position = None, limit = 500):
        request_params = remove_none_values({
            'domainName':domain_name,
            'position':position,
            'limit':limit
        })
        return self.client.get_base_path_mappings(**request_params)
    
    def get_base_path_mappings_with_paginator(self, domain_name):
        return self.get_result_from_paginator('get_base_path_mappings', 'items', domainName=domain_name)

    def get_integration_by_resource_id(self, rest_api_id, resource_id, http_method):
        request_params = remove_none_values({
            'restApiId':rest_api_id,
            'resourceId':resource_id,
            'httpMethod':http_method
        })
        return self.client.get_integration(**request_params)
    
    def get_method_by_resource_id(self, rest_api_id, resource_id, http_method):
        request_params = remove_none_values({
            'restApiId':rest_api_id,
            'resourceId':resource_id,
            'httpMethod':http_method
        })
        return self.client.get_method(**request_params)



import time
import random

from .logger import get_logger

logger = get_logger()

class ExponentialBackoffRetry:
    def __init__(self, max_retries=5, base_delay=0.1, max_delay=5):
        self.max_retries = max_retries
        self.base_delay = base_delay
        self.max_delay = max_delay

    def execute(self, condition_function, execution_function, *args, **kwargs):
        retries = 0

        while retries < self.max_retries:
            try:
                result = execution_function(*args, **kwargs)
                if not condition_function(result):
                    return result

                retries += 1
                if retries > self.max_retries:
                    logger.debug("Max retries reached. Operation failed.")
                    raise Exception("Max retries reached.")
                
                delay = min(self.base_delay * (2 ** retries), self.max_delay)
                delay_with_jitter = delay + random.uniform(0, delay / 2)
                logger.debug(f"Retrying in {delay_with_jitter:.2f} seconds...")
                time.sleep(delay_with_jitter)

            except Exception as e:
                if retries > self.max_retries:
                    logger.debug(f"Operation failed after {retries} retries: {e}")
                    return
                retries += 1
                logger.debug(f"Retry attempt {retries} due to error: {e}")



from . import AWSService


class IdentityStore(AWSService):
    __servicename__ = 'identitystore'

    def list_users(self, identity_store_id):
        return self.client.list_users(IdentityStoreId = identity_store_id)
    
    def list_users_with_paginator(self, identity_store_id):
        request_params = {
            "IdentityStoreId": identity_store_id
        }
        return self.get_result_from_paginator('list_users', 'Users', **request_params)


from . import AWSService
from ..utils.common import remove_none_values
# As "lambda" is a reserved word in python, the file name changed to "lambda_aws"

class Lambda(AWSService):
    __servicename__ = 'lambda'
    
    def get_functions(self, marker = None):
        request_params = remove_none_values({
            'Marker':marker
        })

        return self.client.list_functions(**request_params)

    def list_functions_with_paginator(self):
        return self.get_result_from_paginator('list_functions','Functions')
    
    def get_event_source_mappings(self, marker = None):
        request_params = remove_none_values({
            'Marker':marker
        })
        return self.client.list_event_source_mappings(**request_params)
    
    def list_event_source_mappings_with_paginator(self):
        return self.get_result_from_paginator('list_event_source_mappings', 'EventSourceMappings')


from abc import ABC
import itertools
from boto3.session import Session
import boto3
from ..utils.common import remove_none_values

_available_services = Session().get_available_services()

class AWSService(ABC):
    __servicename__ = None

    def __init__(self, session:Session = None):
        sn = self.__servicename__

        if sn not in _available_services:
            raise Exception(f"Service {sn} is not available")
        
        self.client = session.client(sn) if session is not None else boto3.client(sn)

    def __getattr__(self, name):
        if hasattr(self.client, name):
            return getattr(self.client, name)
        else:
            raise AttributeError(f"'{self.__class__.__name__}' object has no attribute '{name}'")


    def _get_all_with_callback(self, get_list_function, result_token_name, next_token_name, callback_function, *args, next_token = None):
        results = []
        next_token = next_token
        has_next = True
        
        while has_next:
            response = get_list_function(*args, next_token)
            result = response.get(result_token_name)

            if callback_function:
                callback_function(result)
            else:
                if result is not None:
                    results.extend(result)
            next_token = response.get(next_token_name)
            has_next = next_token_name in response

        if not callback_function:
            return results
    
    def _get_all(self, get_list_function, result_token_name, next_token_name, *args):
        return self._get_all_with_callback(get_list_function, result_token_name, next_token_name, None, *args)
    
    def get_result_from_paginator(self, operation_name, result_token, callback_func = None, **kwargs):
        if 'PaginationConfig' not in kwargs:
            kwargs['PaginationConfig'] = {
                'PageSize': 50
            }

        self._gen_iterator = PageResultIterator(self.client, operation_name, kwargs, result_token)

        if callback_func is not None:
            for r in self._gen_iterator:
                callback_func(r)
        else:
            return list(itertools.chain(*self._gen_iterator))
        
    def paginate(self, func, **kwargs):
        if 'PaginationConfig' not in kwargs:
            kwargs['PaginationConfig'] = {
                'PageSize': 50
            }
        
        return PageResultIterator(self.client, getattr(func,"operation_name"), kwargs, getattr(func, "result_token"))
    
class PageResultIterator:
    def __init__(self, client, operation_name, kwargs, result_token):
        self._iterator = client.get_paginator(operation_name).paginate(**remove_none_values(kwargs))
        self._result_token = result_token
    
    def __iter__(self):
        for i in self._iterator:
            yield i.get(self._result_token, [])

def paginateable(operation_name, result_token):
    def decorator(func):
        setattr(func, "operation_name", operation_name)
        setattr(func, "result_token", result_token)
        return func
    return decorator


from typing import Union
from . import AWSService
from ..utils.common import remove_none_values


class S3(AWSService):
    __servicename__ = 's3'
    
    def get_objects_by_bucket_name(self, bucket_name, continuation_token = None):
        request_params = remove_none_values({
            "Bucket":bucket_name,
            'ContinuationToken': continuation_token
        })
        return self.client.list_objects_v2(**request_params)
    
    def list_objects_v2_with_paginator(self, bucket:str):
        return self.get_result_from_paginator('list_objects_v2', 'Contents', Bucket = bucket)

    def get_object(self, s3_path:str) -> bytes:
        bucket_name, key_name = get_bucket_and_key(s3_path)
        response = self.client.get_object(Bucket=bucket_name, Key=key_name)
        return response["Body"].read()
    
    def put_object(self, s3_path:str, body:Union[bytes,str]):
        bucket_name, key_name = get_bucket_and_key(s3_path)
        if isinstance(body,str):
            body = body.encode('utf-8')
        self.client.put_object(Bucket=bucket_name, Key=key_name, Body=body)

def get_bucket_and_key(s3_path:str):
    parts = s3_path.removeprefix('/').split('/',1)
    bucket_name = parts[0]
    key_name = parts[1] if len(parts) > 1 else ""

    return bucket_name, key_name


from datetime import datetime
import time
from . import AWSService
from botocore.utils import (
    datetime2timestamp
)

_CLIENT_REGISTRATION_TYPE = 'public'
_GRANT_TYPE = 'urn:ietf:params:oauth:grant-type:device_code'

class SSO_OIDC(AWSService):
    __servicename__ = 'sso-oidc'

    def register_client(self, client_name = None, client_type = _CLIENT_REGISTRATION_TYPE):
        return self.client.register_client(
            clientName= client_name or ('botocore-client-%s' % int(datetime2timestamp(datetime.now()))),
            clientType=client_type
        )
    
    def start_device_authorization(self, client_id, client_secret, start_url):
        return self.client.start_device_authorization(
            clientId=client_id,
            clientSecret=client_secret,
            startUrl=start_url
        )
    
    def create_token(self, client_id, client_secret, device_code):
        return self.client.create_token(
            grantType= _GRANT_TYPE,
            clientId= client_id,
            clientSecret= client_secret,
            deviceCode= device_code
        )

    def poll_for_token(self, client_id, client_secret, device_code):
        interval = 1

        while True:
            try:
                return self.create_token(client_id, client_secret, device_code)
            except self.client.exceptions.SlowDownException:
                interval += 5
            except self.client.exceptions.AuthorizationPendingException:
                pass
            except self.client.exceptions.ExpiredTokenException as e:
                raise e
            time.sleep(interval)


from . import AWSService, paginateable
from ..utils.common import remove_none_values

class SSO(AWSService):
    __servicename__ = 'sso'

    @paginateable("list_accounts", "accountList")
    def list_accounts(self, accessToken, nextToken = None):
        request_params = remove_none_values({
            'accessToken':accessToken,
            'nextToken':nextToken
        })

        return self.client.list_accounts(**request_params)

    @paginateable("list_account_roles", "roleList")
    def list_account_roles(self, accessToken, accountId, nextToken = None):
        request_params = remove_none_values({
            'accessToken':accessToken,
            'accountId':accountId,
            'nextToken':nextToken
        })

        return self.client.list_account_roles(**request_params)
    
    def get_role_credentials(self, role_name, account_id, sso_access_token):
        return self.client.get_role_credentials(
            roleName=role_name,
            accountId=account_id,
            accessToken=sso_access_token
        ).get('roleCredentials')


from . import AWSService
from ..utils.common import remove_none_values


class APIGatewayV2(AWSService):
    __servicename__ = 'apigatewayv2'

    def get_apis(self, next_token = None):
        request_params = remove_none_values({
            'NextToken': next_token
        })
        return self.client.get_apis(**request_params)
    
    def get_apis_with_paginator(self):
        return self.get_result_from_paginator('get_apis', 'Items')


from . import AWSService
from ..utils.common import remove_none_values


class EventBridge(AWSService):
    __servicename__ = 'events'

    def get_rules(self, next_token = None):
        request_params = remove_none_values({
            'NextToken':next_token
        })
        return self.client.list_rules(**request_params)

    def list_rules_with_paginator(self):
        return self.get_result_from_paginator('list_rules', 'Rules')

    def get_targets_by_rule(self, rule, next_token = None):
        request_params = remove_none_values({
            'Rule':rule,
            'NextToken':next_token
        })
        return self.client.list_targets_by_rule(**request_params)

    def list_targets_by_rule_with_paginator(self, rule):
        return self.get_result_from_paginator('list_targets_by_rule', 'Targets', Rule=rule)



from . import AWSService
from ..utils.common import remove_none_values


class SQS(AWSService):
    __servicename__ = 'sqs'
    
    def get_queues(self, next_token = None):
        request_params = remove_none_values({
            'NextToken':next_token
        })
        return self.client.list_queues(**request_params)

    def list_queues_with_paginator(self):
        return self.get_result_from_paginator('list_queues', 'QueueUrls')
    
    def get_queue_attributes(self, queue_url, attribute_names = ['All']):
        response = self.client.get_queue_attributes(QueueUrl=queue_url,AttributeNames=attribute_names)
        return response.get('Attributes')
    
    def get_queue_url(self, queue_name):
        response = self.client.get_queue_url(QueueName=queue_name)
        return response.get('QueueUrl')
    
    def receive_messages(self, queue_url, number_of_messages, message_attribute_names = ['All'], *, callback_handler):
        while number_of_messages > 0:
            max_number_of_messages = min(number_of_messages, 10)

            response = self.client.receive_message(
                QueueUrl=queue_url,
                AttributeNames=['All'],
                MessageAttributeNames=message_attribute_names,
                MaxNumberOfMessages=max_number_of_messages
            )

            if 'Messages' in response:
                callback_handler(response['Messages'])

            number_of_messages -= max_number_of_messages

    def send_message(self, queue_url, message_body, message_attributes):
        return self.client.send_message(
            QueueUrl = queue_url,
            MessageBody = message_body,
            MessageAttributes = message_attributes
        )
    
    def delete_message(self, queue_url, receipt_handle):
        return self.client.delete_message(
            QueueUrl=queue_url,
            ReceiptHandle=receipt_handle
        )


def chunk_dict(data, max_items_per_chunk):
    chunks = []
    current_chunk = {}
    current_count = 0

    for key, values in data.items():
        for i in range(0, len(values), max_items_per_chunk - current_count):
            chunk_values = values[i:i + (max_items_per_chunk - current_count)]
            if current_chunk.get(key):
                current_chunk[key].extend(chunk_values)
            else:
                current_chunk[key] = chunk_values

            current_count += len(chunk_values)

            if current_count >= max_items_per_chunk:
                chunks.append(current_chunk)
                current_chunk = {}
                current_count = 0

    if current_chunk:
        chunks.append(current_chunk)
    
    return chunks


from . import AWSService
from ..utils.common import remove_none_values


class ResourceGroupsTaggingAPI(AWSService):
    __servicename__ = 'resourcegroupstaggingapi'

    def get_resources(self, pagination_token = None):
        request_params = remove_none_values({
            "PaginationToken":pagination_token
        })
        return self.client.get_resources(**request_params)
    
    def get_resources_with_paginator(self, callback_func = None):
        return self.get_result_from_paginator('get_resources', 'ResourceTagMappingList', callback_func)


from setuptools import setup, find_packages

setup(
    name="botolib",
    version="0.0.29",
    author="Jun Ke",
    author_email="kejun91@gmail.com",
    description="A boto lib that enhances some aws service clients",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    url="https://github.com/kejun91/botolib",
    packages=find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    install_requires=[
        "boto3"
    ],
    python_requires='>=3.9',
)


from . import AWSService
from ..utils.common import remove_none_values


class CloudFormation(AWSService):
    __servicename__ = 'cloudformation'

    def get_stacks(self, next_token = None):
        request_params = remove_none_values({
            'NextToken':next_token
        })
        return self.client.list_stacks(**request_params)
    
    def list_stacks_with_paginator(self):
        return self.get_result_from_paginator('list_stacks', 'StackSummaries')
    
    def get_stack_resources_by_stack_name(self, stack_name, next_token = None):
        request_params = remove_none_values({
            'StackName':stack_name,
            'NextToken':next_token
        })

        return self.client.list_stack_resources(**request_params)
    
    def list_stack_resources_with_paginator(self, stack_name):
        return self.get_result_from_paginator('list_stack_resources','StackResourceSummaries', StackName=stack_name)


from . import AWSService

class STS(AWSService):
    __servicename__ = 'sts'

    def get_caller_identity(self):
        return self.client.get_caller_identity()
    
    def assume_role(self, role_arn, role_session_name):
        return self.client.assume_role(
            RoleArn=role_arn,
            RoleSessionName=role_session_name
        )


from datetime import datetime
from typing import Union, overload
from ..utils.chunk import chunk_dict
from ..utils.retry import ExponentialBackoffRetry
from . import AWSService
from ..utils.common import remove_none_values
from decimal import Decimal
from boto3.dynamodb.types import TypeSerializer, TypeDeserializer
from boto3.dynamodb.conditions import ConditionBase, ConditionExpressionBuilder

class DynamoDB(AWSService):
    __servicename__ = 'dynamodb'

    def batch_get_item(self, keys_with_tables):
        res = self.client.batch_get_item(
            RequestItems=keys_with_tables
        )

        responses = res.get('Responses', {})

        if 'UnprocessedKeys' in res:
            r = self.batch_get_item(res['UnprocessedKeys'])
            for k,v in r.items():
                if k in responses:
                    responses[k].extend(v)
                else:
                    responses[k] = v

        if 'Responses' in res:
            for k,v in responses.items():
                responses[k] = dynamodb_type_to_python_type(v)

        return responses
    
    def get_item(self, table_name:str, primary_key:dict):
        return dynamodb_type_to_python_type(self.client.get_item(
            TableName=table_name,
            Key=python_type_to_dynamodb_type(primary_key)
        ).get('Item'))
    
    def query(self, table_name, index_name, key_condition_expression:ConditionBase, select = None, limit = None, exclusive_start_key = None, filter_expression:ConditionBase = None, selected_attributes = None):
        kwargs = _generate_query_or_scan_kwargs(table_name, index_name, exclusive_start_key, select, limit, key_condition_expression, filter_expression, selected_attributes)
        result = self.client.query(**kwargs)

        if 'Items' in result:
            result['Items'] = dynamodb_type_to_python_type(result['Items'])
        return result
    
    def query_with_paginator(self, table_name, index_name, key_condition_expression:ConditionBase, select = None, filter_expression:ConditionBase = None, selected_attributes = None, callback_handler = None):
        kwargs = _generate_query_or_scan_kwargs(table_name, index_name, None, select, None, key_condition_expression, filter_expression, selected_attributes)

        if callback_handler is not None:
            def convert_and_callback(items):
                callback_handler(dynamodb_type_to_python_type(items))

            return self.get_result_from_paginator('query', 'Items', convert_and_callback, **kwargs)
        else:
            return dynamodb_type_to_python_type(self.get_result_from_paginator('query', 'Items', None, **kwargs))
    
    def scan(self, table_name, index_name = None, filter_expression:ConditionBase = None, exclusive_start_key = None, limit:int = None, select = None, selected_attributes = None):
        kwargs = _generate_query_or_scan_kwargs(table_name, index_name, exclusive_start_key, select, limit, None, filter_expression, selected_attributes)
        result = self.client.scan(**kwargs)

        if 'Items' in result:
            result['Items'] = dynamodb_type_to_python_type(result['Items'])
        return result
    
    def scan_with_paginator(self, table_name, index_name = None, filter_expression:ConditionBase = None, select = None, selected_attributes = None, callback_handler = None):
        kwargs = _generate_query_or_scan_kwargs(table_name, index_name, None, select, None, None, filter_expression, selected_attributes)

        if callback_handler is not None:
            def convert_and_callback(items):
                callback_handler(dynamodb_type_to_python_type(items))

            return self.get_result_from_paginator('scan', 'Items', convert_and_callback, **kwargs)
        else:
            return dynamodb_type_to_python_type(self.get_result_from_paginator('scan', 'Items', None, **kwargs))
    
    def execute_partiql_with_custom_paginator(self, partiql_statement, callback_handler = None, next_token = None):
        return self._get_all_with_callback(self.execute_partiql, 'Items', 'NextToken', callback_handler, partiql_statement, next_token = next_token)
    
    def execute_partiql(self, query_statement, next_token = None):
        request_params = {
            "Statement": query_statement,
            "NextToken": next_token
        }
        result = self.client.execute_statement(**remove_none_values(request_params))
        if 'Items' in result:
            result['Items'] = dynamodb_type_to_python_type(result['Items'])
        return result
    
    def list_tables(self, last_evaluated_table = None):
        request_params = remove_none_values({
            'ExclusiveStartTableName':last_evaluated_table
        })
        return self.client.list_tables(**request_params)
    
    def describe_table(self, table_name):
        res = self.client.describe_table(TableName=table_name)
        return res.get('Table')
    
    def list_tables_with_paginator(self):
        '''
        TableNames, LastEvaluatedTableName, ExclusiveStartTableName
        '''
        return self.get_result_from_paginator('list_tables', 'TableNames')
    
    def put_item(self, table_name, item):
        return self.client.put_item(
            TableName = table_name,
            Item = python_type_to_dynamodb_type(item)
        )
    
    def delete_item(self, table_name, key:dict):
        self.client.delete_item(
            TableName = table_name,
            Key = python_type_to_dynamodb_type(key)
        )

    def batch_write_item(self, put_items:dict = None, delete_keys:dict = None):
        requested_items = {}
        if put_items is not None:
            for tn, items in put_items.items():
                if tn not in requested_items:
                    requested_items[tn] = []
                
                for i in items:
                    requested_items[tn].append({
                        "PutRequest": {
                            "Item": python_type_to_dynamodb_type(i)
                        }
                    })
        
        if delete_keys is not None:
            for tn, keys in delete_keys.items():
                if tn not in requested_items:
                    requested_items[tn] = []

                for k in keys:
                    requested_items[tn].append({
                        "DeleteRequest": {
                            "Key": python_type_to_dynamodb_type(k)
                        }
                    })

        if requested_items != {}:
            #TODO: check bytes of total items for chunk
            for chunk in chunk_dict(requested_items, 25):
                self._batch_write_item_with_retry(chunk)

    def _batch_write_item_with_retry(self, request_items):
        condition_check = lambda req_items: req_items is not None
        execution_func = lambda req_items: self.client.batch_write_item(RequestItems=req_items).get('UnprocessedItems')
        ExponentialBackoffRetry().execute(condition_check, execution_func, request_items)

    def batch_delete_items(self, table_name, keys):
        self.batch_write_item(None, {table_name:keys})

    def batch_put_items(self, table_name, items):
        self.batch_write_item({table_name:items}, None)

    def update_item_with_condition(self, table_name, key, condition_expression:ConditionBase, set_attribute_values:dict, remove_attributes:list = None):

        '''
        Parameters not exposed:
        ReturnConsumedCapacity
        ReturnItemCollectionMetrics
        ReturnValuesOnConditionCheckFailure
        '''

        update_expression, names, values = get_update_expression_attributes(python_type_to_dynamodb_type(set_attribute_values), remove_attributes)

        if condition_expression is not None:
            ce_builder = ConditionExpressionBuilder()
            result = ce_builder.build_expression(condition_expression, False)
            condition_expression = result.condition_expression
            names.update(result.attribute_name_placeholders)
            values.update(python_type_to_dynamodb_type(result.attribute_value_placeholders))
        
        request_params = remove_none_values({
            "TableName": table_name,
            "Key": python_type_to_dynamodb_type(key),
            "UpdateExpression": update_expression,
            "ConditionExpression": condition_expression,
            "ExpressionAttributeNames": names,
            "ExpressionAttributeValues": values,
            "ReturnValues":"UPDATED_NEW"
        })

        return self.client.update_item(**request_params).get("Attributes")
    
    def update_item(self, table_name, key, set_attribute_values:dict, remove_attributes:list = None):
        return self.update_item_with_condition(table_name, key, None, set_attribute_values, remove_attributes)

def get_update_expression_attributes(update_attribute_values, remove_attributes, set_expression_callback = lambda name, value: f"{name} = {value}"):
    update_expression = ''
    i = 0
    expression_attribute_names = {}
    expression_attribute_values = {}
    expressions = []
    if update_attribute_values is not None:
        for n, v in update_attribute_values.items():
            n_alias = f'#name{i}'
            v_alias = f':value{i}'
            i = i + 1
            expressions.append(set_expression_callback(n_alias,v_alias))
            expression_attribute_names.update({n_alias:n})
            expression_attribute_values.update({v_alias:v})

        if len(expressions) > 0:
            update_expression += "SET " + ", ".join(expressions)

    remove_aliases = []
    if remove_attributes is not None:
        for r in remove_attributes:
            n_alias = f'#name{i}'
            remove_aliases.append(n_alias)
            i = i + 1
            expression_attribute_names.update({n_alias:r})

        if len(remove_aliases) > 0:
            update_expression += "REMOVE " + ", ".join(remove_aliases)
    
    return update_expression, expression_attribute_names, expression_attribute_values
    
def _generate_query_or_scan_kwargs(table_name, index_name, exclusive_start_key, select, limit, key_condition_expression, filter_expression, selected_attributes):
    kwargs = {
        "TableName":table_name,
        "IndexName":index_name,
        "ExclusiveStartKey":exclusive_start_key,
        "Select": select,
        "Limit":limit
    }

    ce_builder = ConditionExpressionBuilder()
    expr_attr_names = {}
    expr_attr_values = {}

    if key_condition_expression is not None:
        key_expr_result = ce_builder.build_expression(key_condition_expression, True)
        kwargs['KeyConditionExpression'] = key_expr_result.condition_expression
        expr_attr_names.update(key_expr_result.attribute_name_placeholders)
        expr_attr_values.update(python_type_to_dynamodb_type(key_expr_result.attribute_value_placeholders))

    if filter_expression is not None:
        expr_result = ce_builder.build_expression(filter_expression, False)
        kwargs['FilterExpression'] = expr_result.condition_expression
        expr_attr_names.update(expr_result.attribute_name_placeholders)
        expr_attr_values.update(python_type_to_dynamodb_type(expr_result.attribute_value_placeholders))

    if selected_attributes is not None:
        pe = []
        for i, value in enumerate(selected_attributes):
            a = f'#PE{i}'
            expr_attr_names[a] = value
            pe.append(a)
        kwargs['ProjectionExpression'] = ','.join(pe)
    
    if len(expr_attr_names) > 0:
        kwargs['ExpressionAttributeNames'] = expr_attr_names
    
    if len(expr_attr_values) > 0:
        kwargs['ExpressionAttributeValues'] = expr_attr_values

    return remove_none_values(kwargs)

@overload
def python_type_to_dynamodb_type(item:dict) -> dict: ...

@overload
def python_type_to_dynamodb_type(items:list) -> list: ...

def python_type_to_dynamodb_type(arg:Union[dict, list]) -> Union[dict, list]:
    if arg is None:
        return arg
    elif isinstance(arg, dict):
        dynamodb_item = {}
        serializer = TypeSerializer()
        for k,v in arg.items():
            if isinstance(v, datetime):
                v = v.timestamp()
            if isinstance(v, float):
                v = Decimal(str(v))
            dynamodb_item[k] = serializer.serialize(v)
        return dynamodb_item
    elif isinstance(arg, list):
        return [python_type_to_dynamodb_type(item) for item in arg]
    else:
        raise Exception('arg must be dict or list')
    
@overload
def dynamodb_type_to_python_type(item:dict) -> dict: ...

@overload
def dynamodb_type_to_python_type(items:list) -> list: ...

def dynamodb_type_to_python_type(arg:Union[dict,list]) -> Union[dict,list]:
    if arg is None:
        return arg
    elif isinstance(arg, dict):
        deserializer = TypeDeserializer()
        python_item = {}
        deserializer = TypeDeserializer()
        for k,v in arg.items():
            python_item[k] = deserializer.deserialize(v)
        return python_item
    elif isinstance(arg, list):
        return [dynamodb_type_to_python_type(item) for item in arg]
    else:
        raise Exception('arg must be dict or list')


from . import AWSService
from ..utils.common import remove_none_values


class IoT(AWSService):
    __servicename__ = 'iot'

    def get_thing_describe(self, thing_name):
        return self.client.describe_thing(thingName = thing_name)

    def get_principals_by_thing(self, thing_name, next_token = None):
        request_params = remove_none_values({
            'thingName':thing_name,
            'nextToken':next_token
        })
        
        return self.client.list_thing_principals(**request_params)
    
    def list_thing_principals_with_paginator(self, thing_name:str):
        return self.get_result_from_paginator('list_thing_principals', 'principals', thingName=thing_name)
    
    def get_certificate_describe(self, certificate_id):
        return self.client.describe_certificate(certificateId=certificate_id)



from . import AWSService
from ..utils.common import remove_none_values


class CloudWatch(AWSService):
    __servicename__ = 'cloudwatch'

    def get_metrics(self, next_token = None):
        request_params = remove_none_values({
            'NextToken': next_token
        })
        return self.client.list_metrics(**request_params)
    
    def list_metrics_with_paginator(self):
        return self.get_result_from_paginator('list_metrics', 'Metrics')